package com.ab.abkeystoressl;

import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.BA.Author;
import anywheresoftware.b4a.BA.DependsOn;
import anywheresoftware.b4a.BA.DesignerName;
import anywheresoftware.b4a.BA.ShortName;
import anywheresoftware.b4a.BA.Version;
import anywheresoftware.b4a.objects.collections.List;
import anywheresoftware.b4j.object.ServerWrapper.SslContextFactoryWrapper;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Writer;
import java.net.URI;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
//import java.nio.file.Paths;
//import java.nio.file.StandardCopyOption;
import java.nio.file.StandardOpenOption;
import java.security.GeneralSecurityException;
import java.security.Key;
import java.security.KeyFactory;
import java.security.KeyPair;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.Security;
import java.security.cert.CertificateFactory;
import java.security.spec.PKCS8EncodedKeySpec;
import java.text.SimpleDateFormat;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Base64;
import java.util.Collection;
import java.util.Date;
import java.util.Enumeration;
import java.util.regex.Pattern;

import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.shredzone.acme4j.Account;
import org.shredzone.acme4j.AccountBuilder;
import org.shredzone.acme4j.Authorization;
import org.shredzone.acme4j.Certificate;
import org.shredzone.acme4j.Order;
import org.shredzone.acme4j.Session;
import org.shredzone.acme4j.Status;
import org.shredzone.acme4j.challenge.Challenge;
import org.shredzone.acme4j.challenge.Dns01Challenge;
import org.shredzone.acme4j.challenge.Http01Challenge;
import org.shredzone.acme4j.exception.AcmeException;
import org.shredzone.acme4j.util.CSRBuilder;
import org.shredzone.acme4j.util.KeyPairUtils;
import org.eclipse.jetty.util.log.Log;
import org.eclipse.jetty.util.log.StdErrLog;

@DesignerName("Build 20210315")                                    
@Version(1.08F)                                
@Author("Alain Bailleul")      
@ShortName("ABKeystoreSSL")    
@DependsOn(values={"acme4j-client-2.11", "jose4j-0.7.2", "slf4j-api-1.7.30", "bcprov-jdk15on-168", "bcpkix-jdk15on-168", "acme4j-utils-2.11", "commons-codec-1.15", "annotations-3.0.1"})
public class ABKeystoreSSL {
	private File bananoFile=null;
	private boolean UsesVault = false;
	
	 // File name of the User Key Pair
    private static File USER_KEY_FILE; // = new File("user.key");

    // File name of the Domain Key Pair
    private static File DOMAIN_KEY_FILE; // = new File("domain.key");

    // File name of the CSR
    private static File DOMAIN_CSR_FILE; // = new File("domain.csr");

    // File name of the signed certificate
    private static File DOMAIN_CHAIN_FILE; // = new File("domain-chain.crt");
    
    private static File VAULT = new File("vault"); 

    //Challenge type to be used
    private static final ChallengeType CHALLENGE_TYPE = ChallengeType.HTTP;

    // RSA key size of generated key pairs
    private static final int KEY_SIZE = 2048;

    //private static final Logger LOG = LoggerFactory.getLogger(ClientTest.class);

    private enum ChallengeType {HTTP, DNS}
    
    private Date notBefore = new Date();
    private Date notAfter = new Date();
    
    private static String WWWFolder="";
    
    private static String USER_KEY_FILENAME="user.key";
    private static String DOMAIN_KEY_FILENAME="domain.key";
    private static String DOMAIN_CSR_FILENAME="domain.csr";
    private static String DOMAIN_CHAIN_FILENAME="domain-chain.crt";
    
    /**
     * File name of the User Key Pair. Default user.key
     * 
     * If the file does not exist, a new key pair is generated and saved.
     * 
     * Keep this key pair in a safe place! In a production environment, you will not be
     * able to access your account again if you should lose the key pair.
     */
    public void setUserKeyFile(String name) {
    	//USER_KEY_FILE = new File(name);
    	USER_KEY_FILENAME = name;
    }
    
    /**
     * File name of the Domain Key Pair. Default domain.key
     * 
     * If the file does not exist, a new key pair is generated and saved.
     */
    public void setDomainKeyFile(String name) {
    	//DOMAIN_KEY_FILE = new File(name);
    	DOMAIN_KEY_FILENAME = name;
    }
    
    /**
     * File name of the CSR. Default domain.csr
     */
    public void setDomainCsrFile(String name) {
    	//DOMAIN_CSR_FILE = new File(name);
    	DOMAIN_CSR_FILENAME = name;
    }
    
    /**
     * File name of the signed certificate. Default domain-chain.crt
     */
    public void setDomainChainFile(String name) {
    	//DOMAIN_CHAIN_FILE = new File(name);
    	DOMAIN_CHAIN_FILENAME = name;
    }
    
    /**
     * wwwFolder: needs to be the 'entry point' in your webserver (NOT your B4X app). In my case it was var/www/html and not var/www/)
	 * MUST be accessible on port 80. Is for example your Apache Server.
     */
    public void Initialize(String wwwFolder) {
    	Security.addProvider(new BouncyCastleProvider());
    	WWWFolder = wwwFolder;    	    	
    }
    
    /**
     * Generates a self signed certificate for development purposes.
     * 
     * dir: folder where to put the generated keystore .jks file
     * 
     * keyStoreFileNameNoExtension: name of your keystore file (without the .jks extension)
     * 
     * keyStorePassword: the keyStore password you want to use in your jServer
     * 
     * keyManagerPassword: the keyManager password you want to use in your jServer
     * 
     * vault: a folder name where your key files should be stored/retrieved.
     *        a submap with the keyStoreFileNameNoExtension will be created.   
     *
     */
    public boolean GenerateSelfSignedJKS(String dir, String keyStoreFileNameNoExtension, String keyStorePassword, String keyManagerPassword, String vault) {
    	UsesVault = false;
    	SelfSignedCertificate signedCertificate = new SelfSignedCertificate();
        try {
			signedCertificate.createCertificate(dir, keyStoreFileNameNoExtension, keyStorePassword, keyManagerPassword);			
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		} 
        notBefore = signedCertificate.notBefore;
        notAfter = signedCertificate.notAfter;
        if (!vault.equals("")) {
    		UsesVault = true;
	    	if (!vault.endsWith("/") && !vault.endsWith("\\")) {
	    		vault = vault + "/";
	    	}
	    	vault = vault + keyStoreFileNameNoExtension;
	    	VAULT = new File(vault);
	    	if (!VAULT.exists()) {
	    		VAULT.mkdirs();
	    	}
	    	BufferedWriter bananoWriter = null;
            File temp = new File(vault + "/range.times");
            
    		try {			
    			bananoWriter = new BufferedWriter(new FileWriter(temp));
    			bananoWriter.write(GetNotBefore() + "\n");    			
    			bananoWriter.write(GetNotAfter() + "\n");
    		} catch (Exception e) {
    			e.printStackTrace();
    		} finally {
    			try {
    				bananoWriter.close();				
    			} catch (Exception e) {
    			}
    		}
    		bananoWriter = null;
    		BA.Log("Key files saved in the vault: " + vault);
        } else {
        	BA.LogError("Vault can not be empty!");
    		return false;
        }
        String pattern = "yyyy-MM-dd HH:mm:ss";
		SimpleDateFormat simpleDateFormat = new SimpleDateFormat(pattern);

		String d = simpleDateFormat.format(notAfter);
		
		BA.Log("Expires: " + d);
        return true;
    }
    
    /**
     * Generates a certificate for the given domains. Also takes care for the registration process.
     *
     * domains: Domains to get a common certificate for
     * 
     * dir: folder where to put the generated keystore .jks file
     * 
     * keyStoreFileNameNoExtension: name of your keystore file (without the .jks extension)
     * 
     * keyStorePassword: the keyStore password you want to use in your jServer
     * 
     * keyManagerPassword: the keyManager password you want to use in your jServer
     * 
     * forProduction: if true uses "acme://letsencrypt.org", else "acme://letsencrypt.org/staging"
     * 
     * vault: a folder name where your key files should be stored/retrieved.
     *        a submap with the keyStoreFileNameNoExtension will be created.   
     *        
     * ForceGenerate: if true then the jks file will be generated, ignoring the fact if may not be needed as it is still valid       
     * 
     * TIP: Use the GetNotBefore and GetNotAfter methods right after running fetchCertificate to find the valid date range.
     *      
     *      Those dates are also saved in your vault location as 'range.times'.
     */
    public boolean GenerateJKS(List myDomains, String dir, String keyStoreFileNameNoExtension, String keyStorePassword, String keyManagerPassword, boolean forProduction, String vault, boolean ForceGenerate) { //throws IOException, AcmeException {
    	UsesVault = false;
    	
    	File copyFrom = new File(dir,keyStoreFileNameNoExtension + ".jks");
    	//BA.Log(copyFrom.getAbsolutePath());
    	if (!copyFrom.exists()) {
    		if (!ForceGenerate) {
    			BA.LogError(keyStoreFileNameNoExtension + ".jks not yet created. Force Generate enabled.");
    			ForceGenerate = true;
    		}
    	}    	
    	if (!vault.equals("")) {
    		UsesVault = true;
	    	if (!vault.endsWith("/") && !vault.endsWith("\\")) {
	    		vault = vault + "/";
	    	}
	    	vault = vault + keyStoreFileNameNoExtension;
	    	VAULT = new File(vault);
	    	if (!VAULT.exists()) {
	    		VAULT.mkdirs();
	    	}
	    	File temp = new File(vault + "/range.times");
	    	BufferedReader bananoReader=null;
	    	long notBeforeLng = 0;
	    	long notAfterLng = 0;
	    	
	    	if (temp.exists()) {
	    		try {			
	    			bananoReader = new BufferedReader(new FileReader(temp));
	    			String s1 = bananoReader.readLine();
	    			String s2 = bananoReader.readLine();
	    			notBeforeLng = Long.parseLong(s1);
	    			notAfterLng = Long.parseLong(s2);
	    			notBefore.setTime(notBeforeLng);
	    			notAfter.setTime(notAfterLng);
	    			
	    			Date now = new Date();
		    		Date before = new Date();
		    		
		    		LocalDateTime localDateTime = notBefore.toInstant().atZone(ZoneId.systemDefault()).toLocalDateTime();
		            
		            // plus 60
		            localDateTime = localDateTime.plusDays(60);
		            
		            // convert LocalDateTime to date
		            before = Date.from(localDateTime.atZone(ZoneId.systemDefault()).toInstant());
		    		
		    		if (now.getTime() < before.getTime() && !ForceGenerate) {
		    			BA.Log("Not yet needed to renew the certificate...");
		    			String pattern = "yyyy-MM-dd HH:mm:ss";
		        		SimpleDateFormat simpleDateFormat = new SimpleDateFormat(pattern);

		        		String d = simpleDateFormat.format(notAfter);
		        		
		        		BA.Log("Expires: " + d);
		        		
		        		Log.setLog(new StdErrLog());
		        		
		    			return false;		
		    		}
	    			
	    		} catch (Exception e) {
	    			e.printStackTrace();
	    		} finally {
	    			try {
	    				bananoReader.close();				
	    			} catch (Exception e) {
	    			}
	    		}
	    		bananoReader = null;	
	    		
	    	} 
	    	/*
	    	CopyFile(vault + "/" + USER_KEY_FILE.getName(), USER_KEY_FILE.getName());
	    	CopyFile(vault + "/" + DOMAIN_KEY_FILE.getName(), DOMAIN_KEY_FILE.getName());
	    	CopyFile(vault + "/" + DOMAIN_CSR_FILE.getName(), DOMAIN_CSR_FILE.getName());
	    	CopyFile(vault + "/" + DOMAIN_CHAIN_FILE.getName(), DOMAIN_CHAIN_FILE.getName());
	    	*/
	    	USER_KEY_FILE = new File(vault, USER_KEY_FILENAME);
	    	DOMAIN_KEY_FILE = new File(vault, DOMAIN_KEY_FILENAME);
	    	DOMAIN_CSR_FILE = new File(vault, DOMAIN_CSR_FILENAME);
	    	DOMAIN_CHAIN_FILE = new File(vault, DOMAIN_CHAIN_FILENAME);
    	} else {
    		BA.LogError("Vault can not be empty!");
    		return false;
    	}
    	boolean success=false;
		try {
			success = fetchCertificate(myDomains, dir, keyStoreFileNameNoExtension, keyStorePassword, keyManagerPassword, forProduction);
		} catch (AcmeException | IOException e1) {
						
		}
    	if (success && UsesVault) {
    		/*
    		CopyFile(USER_KEY_FILE.getName(),vault + "/" + USER_KEY_FILE.getName());
        	CopyFile(DOMAIN_KEY_FILE.getName(),vault + "/" + DOMAIN_KEY_FILE.getName());
        	CopyFile(DOMAIN_CSR_FILE.getName(),vault + "/" + DOMAIN_CSR_FILE.getName());
        	CopyFile(DOMAIN_CHAIN_FILE.getName(),vault + "/" + DOMAIN_CHAIN_FILE.getName());
        	*/
        	BufferedWriter bananoWriter = null;
            File temp = new File(vault + "/range.times");
            
    		try {			
    			bananoWriter = new BufferedWriter(new FileWriter(temp));
    			bananoWriter.write(GetNotBefore() + "\n");    			
    			bananoWriter.write(GetNotAfter() + "\n");
    		} catch (Exception e) {
    			e.printStackTrace();
    		} finally {
    			try {
    				bananoWriter.close();				
    			} catch (Exception e) {
    			}
    		}
    		bananoWriter = null;
    		BA.Log("Key files saved in the vault: " + vault);
    		
    		String pattern = "yyyy-MM-dd HH:mm:ss";
    		SimpleDateFormat simpleDateFormat = new SimpleDateFormat(pattern);

    		String d = simpleDateFormat.format(notAfter);
    		
    		BA.Log("Expires: " + d);
    	}
    	Log.setLog(new StdErrLog());
    	return success;
    }
    
    /*
    private void CopyFile(String From, String To) {
    	File copyFrom = new File(From);
    	if (copyFrom.exists()) {
	    	Path from = Paths.get(From);
	    	Path to = Paths.get(To);
	    	try {
				Files.copy(from, to, StandardCopyOption.REPLACE_EXISTING);
			} catch (IOException e) {
				e.printStackTrace();
			}
    	}
    }
    */
    
    /**
     * Reloads the SslConfiguration of the jServer
     * 
     * Based on the reloadSSLConfiguration method written by OliverA
     * More info: https://www.b4x.com/android/forum/threads/access-to-sslfactory-in-jserver-or-implementation-of-sslcontextfactory-reload.128051/#post-802925
     */
    public void ReloadJKS(SslContextFactoryWrapper SslConfiguration) {
        try {
        	SslConfiguration.getObject().reload(sfc -> {});
		} catch (Exception e) {
			e.printStackTrace();
		}
        //sslFactory.reload(sslFactory -> {});
    }
    
    /**
     * Can only be called right after the GenerateJKS() call.
     * Is also saved in your vault location/{keystoreFileNameNoExtension} as 'range.times'.
     */
    public long GetNotBefore() {
    	return notBefore.getTime();
    }
    
    /**
     * Can only be called right after the GenerateJKS() call.
     * Is also saved in your vault location/{keystoreFileNameNoExtension} as 'range.times'.
     */
    public long GetNotAfter() {
    	return notAfter.getTime();
    }
    
    private boolean fetchCertificate(List myDomains, String dir, String keyStoreFileNameNoExtension, String keyStorePassword, String keyManagerPassword, boolean forProduction) throws IOException, AcmeException {
    	Collection<String> domains = new ArrayList<String>();
    	for (int i=0;i<myDomains.getSize();i++) {
    		String s = (String) myDomains.Get(i);
    		domains.add(s);
    	}
    	
        File keyStoreF = new File(dir, keyStoreFileNameNoExtension + ".pkcs12");
        if (keyStoreF.exists()) {
        	keyStoreF.delete();
        }
        
        File keyStore;
        keyStore = new File(dir, keyStoreFileNameNoExtension + ".jks");
		if (keyStore.exists()) {
			keyStore.delete();
		}
    	
        // Load the user key file. If there is no key file, create a new one.
        KeyPair userKeyPair = loadOrCreateUserKeyPair();

        // Create a session for Let's Encrypt.
        // Use "acme://letsencrypt.org" for production server
        String Staging="";
        if (!forProduction) {
        	Staging = "/staging";
        }
        Session session = new Session("acme://letsencrypt.org" + Staging);

        // Get the Account.
        // If there is no account yet, create a new one.
        Account acct = findOrRegisterAccount(session, userKeyPair);

        // Load or create a key pair for the domains. This should not be the userKeyPair!
        KeyPair domainKeyPair = loadOrCreateDomainKeyPair();

        // Order the certificate
        Order order = acct.newOrder().domains(domains).create();

        // Perform all required authorizations
        for (Authorization auth : order.getAuthorizations()) {
            authorize(auth);
        }

        // Generate a CSR for all of the domains, and sign it with the domain key pair.
        CSRBuilder csrb = new CSRBuilder();
        csrb.addDomains(domains);
        csrb.sign(domainKeyPair);
        
        // Write the CSR to a file, for later use.
        try (Writer out = new FileWriter(DOMAIN_CSR_FILE)) {
            csrb.write(out);
        }

        // Order the certificate
        
        order.execute(csrb.getEncoded());       
        
        
        // Wait for the order to complete
        try {
            int attempts = 10;
            while (order.getStatus() != Status.VALID && attempts-- > 0) {
                // Did the order fail?
                if (order.getStatus() == Status.INVALID) {
                    BA.LogError("Order has failed, reason: " + order.getError());
                    return false;
                    //throw new AcmeException("Order failed... Giving up.");
                }

                // Wait for a few seconds
                Thread.sleep(3000L);

                // Then update the status
                order.update();
            }
        } catch (InterruptedException ex) {
        	BA.LogError("interrupted " + ex);
            Thread.currentThread().interrupt();
            return false;
        }

        // Get the certificate
        Certificate certificate = order.getCertificate();

        // Write a combined file containing the certificate and chain.
        try (FileWriter fw = new FileWriter(DOMAIN_CHAIN_FILE)) {
            certificate.writeCertificate(fw);
        }
        
        notBefore = certificate.getCertificate().getNotBefore();
        notAfter = certificate.getCertificate().getNotAfter();
        
        char[] chStore = new char[keyStorePassword.length()];
        for (int i = 0; i < keyStorePassword.length(); i++) { 
            chStore[i] = keyStorePassword.charAt(i); 
        } 
        char[] chManager = new char[keyManagerPassword.length()];
        for (int i = 0; i < keyManagerPassword.length(); i++) { 
            chManager[i] = keyManagerPassword.charAt(i); 
        }
        
        Path cert = DOMAIN_CHAIN_FILE.toPath();
        Path keySt = keyStoreF.toPath();
        
        try {
        	createIdentityStore(cert, DOMAIN_KEY_FILE.toPath(), keySt, chStore);
        	
        	KeyStore kspkcs12 = KeyStore.getInstance("pkcs12");
            KeyStore ksjks = KeyStore.getInstance("jks");

            char[] inphrase = chStore;
            char[] outphrase = chManager;

            kspkcs12.load(new FileInputStream(keyStoreF), inphrase);

            ksjks.load(null, inphrase);
            
            @SuppressWarnings("rawtypes")
			Enumeration eAliases = kspkcs12.aliases();
            int n = 0;
            while (eAliases.hasMoreElements()) {
               String strAlias = (String)eAliases.nextElement();
               BA.LogError("Alias " + n++ + ": " + strAlias);

               if (kspkcs12.isKeyEntry(strAlias)) {
                  BA.LogError("Adding key for alias " + strAlias);
                  Key key = kspkcs12.getKey(strAlias, inphrase);

                  java.security.cert.Certificate[] chain = kspkcs12.getCertificateChain(strAlias);

                  ksjks.setKeyEntry(strAlias, key, outphrase, chain);
               }
            }

            OutputStream out = new FileOutputStream(keyStore);
            ksjks.store(out, inphrase);
            out.close();			
		} catch (GeneralSecurityException e) {
			e.printStackTrace();
			return false;
		}
        
        if (bananoFile!=null) {
			  if (bananoFile.exists()) {
				  BA.Log("Removing temporary challenge file: " + bananoFile.getAbsolutePath());
				  bananoFile.delete();
			  }
		}
        if (keyStoreF!=null) {
        	if (keyStoreF.exists()) {
        		keyStoreF.delete();
        	}
        }
        
        BA.Log("Success! The certificate for domains has been generated!");
        //BA.Log("Certificate url: " + certificate.getLocation());
        
        BA.Log("Certificate location: " + keyStore.getAbsolutePath());
        BA.LogError("Keep the " + USER_KEY_FILE.getName() + " key pair file in a safe place! In a production environment, you will not be able to access your account again if you should lose the key pair.");
        
        
        
        return true;
    }
   
    /**
     * Loads a user key pair from {@link #USER_KEY_FILE}. If the file does not exist, a
     * new key pair is generated and saved.
     * <p>
     * Keep this key pair in a safe place! In a production environment, you will not be
     * able to access your account again if you should lose the key pair.
     *
     * @return User's {@link KeyPair}.
     */
    private KeyPair loadOrCreateUserKeyPair() throws IOException {
        if (USER_KEY_FILE.exists()) {
            // If there is a key file, read it
            try (FileReader fr = new FileReader(USER_KEY_FILE)) {
                return KeyPairUtils.readKeyPair(fr);
            }

        } else {
            // If there is none, create a new key pair and save it
            KeyPair userKeyPair = KeyPairUtils.createKeyPair(KEY_SIZE);
            try (FileWriter fw = new FileWriter(USER_KEY_FILE)) {
                KeyPairUtils.writeKeyPair(userKeyPair, fw);
            }
            return userKeyPair;
        }
    }

    /**
     * Loads a domain key pair from {@link #DOMAIN_KEY_FILE}. If the file does not exist,
     * a new key pair is generated and saved.
     *
     * @return Domain {@link KeyPair}.
     */
    private KeyPair loadOrCreateDomainKeyPair() throws IOException {
        if (DOMAIN_KEY_FILE.exists()) {
            try (FileReader fr = new FileReader(DOMAIN_KEY_FILE)) {
                return KeyPairUtils.readKeyPair(fr);
            }
        } else {
            KeyPair domainKeyPair = KeyPairUtils.createKeyPair(KEY_SIZE);
            try (FileWriter fw = new FileWriter(DOMAIN_KEY_FILE)) {
                KeyPairUtils.writeKeyPair(domainKeyPair, fw);
            }
            return domainKeyPair;
        }
    }

    /**
     * Finds your {@link Account} at the ACME server. It will be found by your user's
     * public key. If your key is not known to the server yet, a new account will be
     * created.
     * <p>
     * This is a simple way of finding your {@link Account}. A better way is to get the
     * URL of your new account with {@link Account#getLocation()} and store it somewhere.
     * If you need to get access to your account later, reconnect to it via {@link
     * Session#login(URL, KeyPair)} by using the stored location.
     *
     * @param session
     *         {@link Session} to bind with
     * @return {@link Account}
     */
    private Account findOrRegisterAccount(Session session, KeyPair accountKey) throws AcmeException {
        // Ask the user to accept the TOS, if server provides us with a link.
        URI tos = session.getMetadata().getTermsOfService();
        if (tos != null) {
            //acceptAgreement(tos);
        }

        Account account = new AccountBuilder()
                .agreeToTermsOfService()
                .useKeyPair(accountKey)
                .create(session);
        BA.Log("Registered a user, URL: " + account.getLocation());

        return account;
    }

    /**
     * Authorize a domain. It will be associated with your account, so you will be able to
     * retrieve a signed certificate for the domain later.
     *
     * @param auth
     *         {@link Authorization} to perform
     */
    private void authorize(Authorization auth) throws AcmeException {
    	BA.Log("Authorization for domain " + auth.getIdentifier().getDomain());

        // The authorization is already valid. No need to process a challenge.
        if (auth.getStatus() == Status.VALID) {
        	BA.LogError("The authorization is already valid. No need to process a challenge.");
            return;
        }

        // Find the desired challenge and prepare it.
        Challenge challenge = null;
        switch (CHALLENGE_TYPE) {
            case HTTP:
                challenge = httpChallenge(auth);
                break;

            case DNS:
                challenge = dnsChallenge(auth);
                break;
        }

        if (challenge == null) {
        	BA.LogError("No challenge found.");
            throw new AcmeException("No challenge found.");
        }

        // If the challenge is already verified, there's no need to execute it again.
        if (challenge.getStatus() == Status.VALID) {
        	BA.LogError("Challenge already verified, there's no need to execute it again.");
            return;
        }

        // Now trigger the challenge.
        challenge.trigger();

        // Poll for the challenge to complete.
        try {
            int attempts = 10;
            while (challenge.getStatus() != Status.VALID && attempts-- > 0) {
                // Did the authorization fail?
                if (challenge.getStatus() == Status.INVALID) {
                	BA.LogError("Challenge has failed, reason: " + challenge.getError());
                	return;                    
                }

                // Wait for a few seconds
                Thread.sleep(3000L);

                // Then update the status
                challenge.update();
            }
        } catch (InterruptedException ex) {
        	BA.LogError("interrupted" + ex);
            Thread.currentThread().interrupt();
        }

        // All reattempts are used up and there is still no valid authorization?
        if (challenge.getStatus() != Status.VALID) {
        	BA.LogError("Failed to pass the challenge for domain " + auth.getIdentifier().getDomain() + ", ... Giving up.");
        	return;
        }

        BA.Log("Challenge has been completed.\nYou can remove the resource again now.");
    }

    /**
     * Prepares a HTTP challenge.
     * <p>
     * The verification of this challenge expects a file with a certain content to be
     * reachable at a given path under the domain to be tested.
     * <p>
     * This example outputs instructions that need to be executed manually. In a
     * production environment, you would rather generate this file automatically, or maybe
     * use a servlet that returns {@link Http01Challenge#getAuthorization()}.
     *
     * @param auth
     *         {@link Authorization} to find the challenge in
     * @return {@link Challenge} to verify
     */
    protected Challenge httpChallenge(Authorization auth) throws AcmeException {
        // Find a single http-01 challenge
        Http01Challenge challenge = auth.findChallenge(Http01Challenge.class);
        if (challenge == null) {
        	BA.LogError("Found no " + Http01Challenge.TYPE + " challenge, don't know what to do...");
        	return challenge;            
        }

        BA.Log("Reachable at: http://" + auth.getIdentifier().getDomain() + "/.well-known/acme-challenge/" + challenge.getToken());
        BA.Log("File name: " + challenge.getToken());
        BA.Log("Content: " + challenge.getAuthorization());
        
        BufferedWriter bananoWriter = null;
        bananoFile = new File(WWWFolder + "/.well-known/acme-challenge/");
        if (!bananoFile.exists()) {
        	bananoFile.mkdirs();
        }
        bananoFile = new File(WWWFolder + "/.well-known/acme-challenge/" + challenge.getToken());
        BA.Log("Creating temporary challenge file: " + bananoFile.getAbsolutePath());
		
		try {			
			bananoWriter = new BufferedWriter(new FileWriter(bananoFile));
			bananoWriter.write(challenge.getAuthorization());
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			try {
				bananoWriter.close();				
			} catch (Exception e) {
			}
		}
		bananoWriter = null;

        return challenge;
    }

    /**
     * Prepares a DNS challenge.
     * <p>
     * The verification of this challenge expects a TXT record with a certain content.
     * <p>
     * This example outputs instructions that need to be executed manually. In a
     * production environment, you would rather configure your DNS automatically.
     *
     * @param auth
     *         {@link Authorization} to find the challenge in
     * @return {@link Challenge} to verify
     */
    protected Challenge dnsChallenge(Authorization auth) throws AcmeException {
        // Find a single dns-01 challenge
        Dns01Challenge challenge = auth.findChallenge(Dns01Challenge.TYPE);
        if (challenge == null) {
        	BA.LogError("Found no " + Dns01Challenge.TYPE + " challenge, don't know what to do...");
        	return challenge;
        }

        // Output the challenge, wait for acknowledge...
        BA.Log("Please create a TXT record:");
        BA.Log("_acme-challenge." + auth.getIdentifier().getDomain() + ". IN TXT " + challenge.getDigest());
        
        /*
        TODO auto create the .TXT file possible?        
        */

        return challenge;
    }
    
    protected void createTrustStore(Path certificate, Path keystore, char[] password) throws IOException, GeneralSecurityException {
        CertificateFactory cf = CertificateFactory.getInstance("X.509");
        java.security.cert.Certificate root;
        try (InputStream is = Files.newInputStream(certificate)) {
            root = cf.generateCertificate(is);
        }
        KeyStore pkcs12 = KeyStore.getInstance("PKCS12");
        pkcs12.load(null, null);
        pkcs12.setCertificateEntry("root", root);
        try (OutputStream os = Files.newOutputStream(keystore, StandardOpenOption.CREATE_NEW)) {
            pkcs12.store(os, password);
        }
    }

    private static final byte[] HEADER = "-----".getBytes(StandardCharsets.US_ASCII);

    protected void createIdentityStore(Path certificate, Path key, Path keystore, char[] password) throws IOException, GeneralSecurityException {
        byte[] pkcs8 = decode(Files.readAllBytes(key));
        KeyFactory kf = KeyFactory.getInstance("RSA");
        PrivateKey pvt = kf.generatePrivate(new PKCS8EncodedKeySpec(pkcs8));
        CertificateFactory cf = CertificateFactory.getInstance("X.509");
        java.security.cert.Certificate pub;
        try (InputStream is = Files.newInputStream(certificate)) {
            pub = cf.generateCertificate(is);
        }
        KeyStore pkcs12 = KeyStore.getInstance("PKCS12");
        pkcs12.load(null, null);
        pkcs12.setKeyEntry("identity", pvt, password, new java.security.cert.Certificate[] { pub });
        try (OutputStream s = Files.newOutputStream(keystore, StandardOpenOption.CREATE_NEW)) {
            pkcs12.store(s, password);
        }
    }

    private static byte[] decode(byte[] raw) {
        if (!Arrays.equals(Arrays.copyOfRange(raw, 0, HEADER.length), HEADER)) return raw;
        CharBuffer pem = StandardCharsets.US_ASCII.decode(ByteBuffer.wrap(raw));
        String[] lines = Pattern.compile("\\R").split(pem);
        String[] body = Arrays.copyOfRange(lines, 1, lines.length - 1);
        return Base64.getDecoder().decode(String.join("", body));
    }
    
}
