package com.ab.abkeystoressl;

import anywheresoftware.b4a.BA.Hide;
import java.io.File;
import java.io.FileOutputStream;
import java.math.BigInteger;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.SecureRandom;
import java.security.Security;
import java.security.cert.X509Certificate;
import java.util.Date;

import org.bouncycastle.jce.X509Principal;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.x509.X509V3CertificateGenerator;


@SuppressWarnings("deprecation")
@Hide
public class SelfSignedCertificate {
	 private static String CERTIFICATE_ALIAS = "identity";
	 private static String CERTIFICATE_ALGORITHM = "RSA";
	 private static String CERTIFICATE_DN = "CN=cn, O=o, L=L, ST=il, C= c";
	 private static String CERTIFICATE_NAME = "keystore.jks";
	 private static final int CERTIFICATE_BITS = 2048;
	 private static String KEYSTORE_PWD="";
	 private static String KEYMANAGER_PWD="";
	 
	 public Date notBefore = new Date();
	 public Date notAfter = new Date();
	    
	 static {
	        // adds the Bouncy castle provider to java security
	        Security.addProvider(new BouncyCastleProvider());
	 }
	 
	 public X509Certificate createCertificate(String dir, String keyStoreFileNameNoExtension, String keyStorePassword, String keyManagerPassword) throws Exception{
		    KEYSTORE_PWD = keyStorePassword;
		    KEYMANAGER_PWD = keyManagerPassword;
		    CERTIFICATE_NAME = keyStoreFileNameNoExtension + ".jks";
	        X509Certificate cert = null;
	        KeyPairGenerator keyPairGenerator = KeyPairGenerator.getInstance(CERTIFICATE_ALGORITHM);
	        keyPairGenerator.initialize(CERTIFICATE_BITS, new SecureRandom());
	        KeyPair keyPair = keyPairGenerator.generateKeyPair();

	        // GENERATE THE X509 CERTIFICATE
	        X509V3CertificateGenerator v3CertGen =  new X509V3CertificateGenerator();
	        v3CertGen.setSerialNumber(BigInteger.valueOf(System.currentTimeMillis()));
	        v3CertGen.setIssuerDN(new X509Principal(CERTIFICATE_DN));
	        v3CertGen.setNotBefore(new Date(System.currentTimeMillis() - 1000L * 60 * 60 * 24));
	        v3CertGen.setNotAfter(new Date(System.currentTimeMillis() + (1000L * 60 * 60 * 24 * 365*10)));
	        v3CertGen.setSubjectDN(new X509Principal(CERTIFICATE_DN));
	        v3CertGen.setPublicKey(keyPair.getPublic());
	        v3CertGen.setSignatureAlgorithm("SHA256WithRSAEncryption");
	        cert = v3CertGen.generateX509Certificate(keyPair.getPrivate());
	        notBefore = cert.getNotBefore();
	        notAfter = cert.getNotAfter();
	        saveCert(dir, cert,keyPair.getPrivate());
	        return cert;
	 }

	 private void saveCert(String dir, X509Certificate cert, PrivateKey key) throws Exception {
	        KeyStore keyStore = KeyStore.getInstance("JKS");    
	        keyStore.load(null, null);
	        keyStore.setKeyEntry(CERTIFICATE_ALIAS, key, KEYMANAGER_PWD.toCharArray(),  new java.security.cert.Certificate[]{cert});
	        File file = new File(dir, CERTIFICATE_NAME);
	        keyStore.store( new FileOutputStream(file), KEYSTORE_PWD.toCharArray() );
	 }
}
